%%% -*- coding: latin-1 -*-

%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Etvs Lornd University.
%%% Portions created  by Etvs  Lornd University are  Copyright 2007-2011,
%%% Etvs Lornd University. All Rights Reserved.

%%% @doc Standalone static analyser for dynamic call constructs.
%%%
%%% @author Dniel Horpcsi <daniel_h@inf.elte.hu>

-module(refanal_dynfun_p).
-vsn("$Rev: 7473 $").

-export([analyse/0, clean/0]).

-include("core.hrl").
-include_lib("referl_lib/include/lib_export.hrl").

-define(CallAnal, refcore_callanal).

-define(timeout, 10000).

-type gnode(_) :: tuple().

%%  3: exactly 3
%% -1: unknown
%% -4: at least 3
-type dyn_arity() :: integer().

-record(funref, {expr  :: gnode(expr),
                 id    :: {?CallAnal:id(), ?CallAnal:id()},
                 arity :: dyn_arity() | [dyn_arity()]
                }).

-define(links, [{dynfuneref,back}, {dynfunlref,back},
                {ambfuneref,back}, {ambfunlref,back},
                {may_be, back},
                {dyncall, back}, {ambcall, back}]).

-define(out(S),    io:format("~n" ++ S)   ).
-define(out(S, A), io:format("~n" ++ S, A)).

%% =============================================================================
%% Cleaning

%% TODO how to delete modref links?
clean() ->
    [clean(Fun) || Fun <- ?Query:exec(?Query:seq([?Mod:all(), ?Mod:locals()]))],
    ?NodeSync:clean(),
    ?FunProp:reset().
clean(Fun) ->
    [clean(Fun, Link) || Link <- ?links].
clean(Fun, {Link,back}) ->
    [?Graph:rmlink(Expr, Link, Fun) || Expr <- ?Graph:path(Fun, [{Link,back}])].

%% =============================================================================
%% Analysis process

analyse() ->
    try
        anal_dyncalls(
          lookup_dyncalls(
            filter_dyncalls(
              all_call_exprs())))
    catch
        no_calls ->
            ?out("No function calls found in the database.");
        no_dynamic ->
            ?out("No dynamic calls found in the database.")
    end,
    ?out("~nAnalysis completed.~n").

mnesia_get_applications() ->
    Match = {expr, '$1', #expr{type=application, _='_'}, '_'},
    mnesia:select(expr, [{Match, [], ['$1']}]).

all_call_exprs() ->
    {atomic, Res} = mnesia:transaction(fun mnesia_get_applications/0),
    lists:map(fun(N) -> {'$gn', expr, N} end, Res).

filter_dyncalls([]) -> throw(no_calls);
filter_dyncalls(AllCalls) ->
    ?out("~p function calls found in the database.", [length(AllCalls)]),
    ?out("Looking for dynamic calls..."),
    DynCalls = [{Type, Expr} || Expr <- AllCalls,
                                Type <- [call_type(Expr)],
                                Type /= static],
    DynCalls.

lookup_dyncalls([]) -> throw(no_dynamic);
lookup_dyncalls(DynCalls) ->
    ?out("~p function calls seem to be dynamic.", [length(DynCalls)]),
    ?out("Looking up dynamic calls...", []),
    S1 = "\rLooking up dynamic calls... (~5w/~5w)",
    S2 = "~nIdentification of ~p dynamic calls timed out.~n",
    Info1 = parallel(fun dynamic_call/2, DynCalls,
                     fun(X, Y) -> io:format(S1, [X, Y]) end,
                     fun(X) -> io:format(S2, [X]) end),
    Info = lists:flatten(Info1),
    ?out("~p dynamic calls identified in the database.", [length(Info)]),
    Info.

anal_dyncalls(Info) ->
    ?out("Analysing dynamic calls..."),
    lists:mapfoldl(
      fun(X, I) -> 
              store(X),
              update_funprop(X#funref.expr),
              io:format("\rAnalysing dynamic calls... (~5w/~5w)", [I, length(Info)]),
              {ok, I+1}
      end, 1, Info).

parallel(Fun, Xs, O1, O2) ->
    Self = self(),
    Pids = [spawn(fun() -> Fun(Self, X) end) || X <- Xs],
    receive_results(Pids, [], length(Pids), O1, O2).
receive_results([], Results, _C, _, _) ->
    Results;
receive_results(Pids, Results, C, O1, O2) ->
    receive {Pid, Rep} ->
            exit(Pid, normal),
            O1(C-length(Pids)+1, C),
            receive_results(Pids -- [Pid], Results ++ [Rep], C, O1, O2)
    after ?timeout ->
            [exit(Pid, kill) || Pid <- Pids],
            O2(length(Pids)),
            receive_results([], Results, C, O1, O2)
    end.

%% =============================================================================
%% Storing dyamic calls into the database

%% TODO: eliminating the synchronisation?

store(#funref{expr = Expr, id = ID, arity = Arities}) when is_list(Arities) ->
    [store(#funref{expr = Expr, id = ID, arity = Arity}) || Arity <- Arities];

store(#funref{expr = Expr, id = ID, arity = Arity}) ->
    case ID of

        %% Unambiguous identifiers / possibly opaque arity

        {{_ModRef, ModName}, {_FunRef, FunName}} ->
            %%?NodeSync:add_ref(module, {ref, ModRef}, ModName),
            ?NodeSync:add_ref(func, {dyneref, Expr}, {ModName, {FunName, Arity}});

        %% Dynamic call with opaque module name

        {undefined, {_FunRef, FunName}} ->
            ?NodeSync:add_ref(func, {dyneref, Expr}, {-1, {FunName, Arity}});

        %% Dynamic call with opaque function name

        {{_ModRef, ModName}, undefined} ->
            %%?NodeSync:add_ref(module, {ref, ModRef}, ModName),
            ?NodeSync:add_ref(func, {dyneref, Expr}, {ModName, {-1, Arity}});

        %% Fully ambiguous calls --- skipped!

        undefined              -> ok;
        {undefined, undefined} -> ok;

        %% Multiple references handled one by one (recursion!)

        {ModIDs, FunIDs} when is_list(ModIDs); is_list(FunIDs) ->
            [store(#funref{expr = Expr, id = {M, F}, arity = Arity}) ||
                M <- lists:flatten([ModIDs]), F <- lists:flatten([FunIDs])];
        IDs when is_list(IDs) ->
            [store(#funref{expr = Expr, id = I, arity = Arity}) || I <- IDs]
    end.

%% =============================================================================

call_type(CallExpr) ->
    case ?Graph:path(CallExpr, [funlref]) ++ ?Graph:path(CallExpr, [funeref]) of
        [] -> dynamic_mfa;
        [Fun] ->
            [Mod] = ?Graph:path(Fun, ?Fun:module()),
            case {?Graph:data(Mod), ?Graph:data(Fun)} of
                {#module{name=erlang}, #func{name=apply, arity=3}} -> apply;
                _ -> static
            end;
        _ -> static
    end.

lookup_funID(C, #expr{type=infix_expr, value=':'}) ->
    [Mod, Fun] = ?Query:exec(C, ?Expr:children()),
    ?CallAnal:lookup_IDs(Mod, Fun, undefined);
%% TODO include the dynamic funexpr analysis from the good old dynfun
%lookup_funID(N, #expr{}) -> lookup_funexpr_via_dataflow(N);
lookup_funID(_, _) -> undefined.

dynamic_call(Parent, {dynamic_mfa, CallExpr}) ->
    [{esub, FunId}, {esub, ArgLst}] = ?Syn:children(CallExpr),
    Args = [A || {esub, A} <- ?Syn:children(ArgLst)],
    I = [#funref{expr = CallExpr, arity = length(Args),
                 id = lookup_funID(FunId, ?Graph:data(FunId))}],
    Parent ! {self(), I};
dynamic_call(Parent, {apply, CallExpr}) ->
    Parent ! {self(), apply_call(CallExpr)}.

%% TODO erlang:apply/2
apply_call(CallExpr) ->
    [_, ApplyArgs] = ?Query:exec(CallExpr, ?Expr:children()),
    [ApplyModRef, ApplyFunRef, ListOfArgs] = ?Query:exec(ApplyArgs, ?Expr:children()),
    ID = ?CallAnal:lookup_IDs(ApplyModRef, ApplyFunRef, undefined),
    Arity = case ?CallAnal:listcons_length(ListOfArgs) of
                incalculable -> -1;
                {I}          -> -1*I-1;
                I            -> I
            end,
    [#funref{expr = CallExpr, arity = Arity, id = ID}].

%% =============================================================================
%% Copied from refanal_fun and slightly modified

update_funprop(Node) ->
    case ?Graph:data(Node) of
        #form{type=func} -> update_form_funprop(Node);
        #form{} -> ok;
        _D ->
            case ?Syn:parent(Node) of
                [{_,P}] -> update_funprop(P);
                _ -> ok
            end
    end.

update_form_funprop(Form) ->
    %% TODO: maybe ?NodeSync:get_node support
    case ?Graph:path(Form, [fundef]) of
        [Fun] -> ?FunProp:update(Fun, funprops(Form));
        []    -> ok
    end.

funprops(Form) -> funprops(Form, {true, []}).

funprops(Node, Props) ->
    lists:foldl(fun funprops/2, node_funprop(Node, ?Graph:data(Node), Props),
                [Child || {_, Child} <- ?Syn:children(Node)]).

node_funprop(Node, #expr{type=Type}, {Pure, Calls}) ->
    {Pure andalso Type /= send_expr andalso Type /= receive_expr,
     if
         Type == application -> add_calls(Node, Calls);
         true -> Calls
     end};

node_funprop(_, _, Props) -> Props.

add_calls(Node, Calls) ->
    FunCalls = [{funcall, F} || {T, F} <- ?Graph:links(Node), lists:member(T, [funeref, funlref])],
    DynCalls = [{dyncall, F} || {T, F} <- ?Graph:links(Node), lists:member(T, [dynfuneref, dynfunlref])],
    AmbCalls = [{ambcall, F} || {T, F1} <- ?Graph:links(Node), lists:member(T, [ambfuneref, ambfunlref]),
                                F <- ?Graph:path(F1, [may_be])],
    ordsets:union(Calls, lists:usort(FunCalls++DynCalls++AmbCalls)).
