%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös  Loránd University are  Copyright 2007-2009,
%%% Eötvös Loránd University. All Rights Reserved.

%%% @doc Cyclic depedency examination in function level.
%%% Draws out the dependency graph. 
%%% At first run a graph is used, which is stored into a dets table for further
%%% and faster use.
%%% <br/> X function is
%%% dependent from Y function if there is a Y function call in the
%%% definition of X. (X -> Y)<br/> For example, X -> Y -> Z -> X is a
%%% cycle.
%%% @author Kinga Szava <guca04@gmail.com>

-module(refusr_cyclic_fun).
-vsn("$Rev: 7422 $ ").
-export([print_cycle/0, check_cycle/0, check_function/1]).

-export([draw/0, draw/1, draw_cycles/0, draw_cycles/1]).

-export([build_table/3]).

-export([funcalls/1]).

-include("user.hrl").

-define(LibRel, refusr_lib_relations).

%%% ============================================================================
%%% Building the dets table

%% @spec build_table(Name, Args, Data) -> ok
%% @doc Builds up the proper dets table from a built digraph.
%% Only for development use. (Data argument is only for the uniformity.)
build_table(Name, [], _)->
    Graph = digraph:new(),
    try
        build_graph(Graph),
        Cycles = cycles(Graph, digraph:vertices(Graph), [], []),
	?LibRel:build_paths(Graph, Name, Cycles, func)
    after
        digraph:delete(Graph)
    end;
build_table(Name, Args, _)->
    Graph = digraph:new(),
    try
        build_subgraph(Graph, Args),
        Cycles = cycles(Graph, digraph:vertices(Graph), [], []),
	?LibRel:build_paths(Graph, Name, Cycles, func)
    after
        digraph:delete(Graph)
    end.

%%===========================================================================
%% @spec check_cycle() -> {ok, no_cyclic_dependency} | {integer(), list()}
%% @doc Builds a directed graph, checks for cycles in function level. If
%% cycles are found then the following tuple is the result: {Number of
%% cycles, Cycle list}.
check_cycle()->
  {_, Res} = check_function([]),
  {integer_to_list(length(Res)) ++  " cycle(s)", Res}.

%% @spec check_function(Function)-> Result | {error, bad_function_name}
%% | erlang:error(error, Args)
%% Function = node() | string()
%% Result = {Bool | list()}
%% Bool = true | false
%% @doc Creates a subgraph from the given function as a starting node
%% and checks for cycles. The result is a tuple, where the first element
%% is true/false depending on whether there's cycle ot not from the given node. 
%% The second is the a gnode list in which the nodes in the cycle are indicated.
%% <br/>The function can be given as a graph node or  in the following
%% order and combination:<br/> "module:name/arity" (eg.: "erlang:hd/1").
check_function(Function)->
   output(?LibRel:invest_table(func, [{type, check}, {gnode, Function}])).

output([])->
	{false, []};
output(Res)->
	{true, Res}.
   
%% @spec print_cycle() -> ok
%% @doc Prints the result of check_cycle/0 to the standard output.
print_cycle()->
   print_cycle([]).

print_cycle(FromNode)->
   ?LibRel:invest_table(func, [{type, print}, {gnode, FromNode}]).

%% @spec draw()-> {ok, no_cyclic_dependency} | {integer(), list()}
%% @doc Equivalent to draw(""), prints out the entire graph. Output file is dep_func.dot.
draw()->
  draw([{gnode, []}]).


%% @spec draw(Options::proplists())-> {ok, no_cyclic_dependency} | {integer(), list()} |
%% {error, bad_function_name} | erlang:error(error, Args)
%% @doc Creates a subgraph drawing from the directed graph from a given function as a starting node.
%% The function can be given as a graph node or  in the following
%% order and combination: "module:name/arity" (eg.: "erlang:hd/1").<br/>
%% Options:
%% ```
%% {dot, DotFileName::string()}
%%
%% '''
%%
%% ```
%% {gnode, Node::node() | string()}
%% '''
%% The file's name: dep_func_<i>module_function_arity</i>.dot or it can be given by
%% the user with a file name, then the .dot file will be in the the ./dep_files/ directory, 
%% or with absolute path.
%% <br/> The indicators are the following:<br/>
%% - Red edge: cycle edge<br/>
%% - Dashed edge: funcall edge, a function calls, depends on another<br/>
%% - Purple box: module<br/>
%% - Black hexagon: function
draw(Options) when not is_list(Options) ->
	draw([{gnode, Options}]);
draw(Options)->
    case proplists:is_defined(gnode, Options) of
	true ->
           ?LibRel:invest_table(func,[{type, draw} | Options]);
	false ->
		case proplists:is_defined(dot, Options) of
			true ->
			    ?LibRel:invest_table(func,[{type, draw} | Options]);
			false ->
				io:format("Neither dot nor gnode key given properly.~n"),
				draw([{gnode, Options}])
		end
    end.

%% @spec draw_cycles()-> {ok, no_cyclic_dependency} | {integer(), list()}
%% @doc Prints out a subgraph which contains the cycle(s).
%% Unless cycles exist, calls {@link draw/0}.
%% Output file is dep_func_cycles.dot.
draw_cycles()->
    draw_cycles([{gnode, []}]).

%% @spec draw_cycles(PropList::proplists())-> {ok, no_cyclic_dependency} | {integer(), list()} |
%% {error, bad_function_name} | erlang:error(error, Args)
%% @doc Prints out a subgraph which contains the cycle(s) from the given node.
%% Unless cycles exist, calls {@link draw/0}.
%% Output file is dep_func_cycles.dot or user_defined. <br/>
%% Options:
%% ```
%% {dot, DotFileName::string()}
%%
%% '''
%%
%% ```
%% {gnode, Nodes}
%% Nodes = Node | NodeList::lists()
%% Node = string() | gnode()
%% '''
draw_cycles(Options)->
     ?LibRel:invest_table(func,[{type, cycles} | Options]).



%%% ============================================================================
%%% Graph queries
get_all_functions()->
    ?Query:exec(?Query:seq(?Mod:all(), ?Mod:locals())).

get_funcall(Functions)->
    [{Fun, Callees} || Fun <- Functions ,
                       Callees <- funcalls(Fun),
                       Callees /= []].

get_func_mod(Fun)->
    ?Query:exec1(Fun, ?Fun:module(), mod_not_found).

funcalls({_, Type, _}) when Type /= func->
	io:format("Error, bad argument given, a function type gnode should be the argument!"),
	throw({error, written});
funcalls(Fun)->
	[diff_edge_calls(Fun, [funcall, dyncall, ambcall, may_be], [])].

diff_edge_calls(_, [], Result)->
	Result;
diff_edge_calls(Node, [Hd | Rest], Result)->
	diff_edge_calls(Node, Rest, 
		 Result ++ ?Graph:path(Node, [Hd])).

%%% ============================================================================
%%% Building a digraph
build_graph(Graph)-> build_vertex(Graph, get_funcall(get_all_functions())).

build_vertex(_, [])-> ok;
build_vertex(Graph, [{Fun, Callees} | Rest]) ->
    check_subgraph(Graph, Fun),
    [begin
         check_subgraph(Graph, Callee),
         check_edge(Graph, Fun, Callee, "called")
     end
     || Callee <- Callees],
    build_vertex(Graph, Rest).

check_subgraph(Graph, Fun) ->
    Mod = get_func_mod(Fun),
    check_vertex(Graph, Mod),
    check_vertex(Graph, Fun),
    check_edge(Graph, Mod, Fun, "").

check_vertex(Graph, Vertex)->
    case digraph:vertex(Graph, Vertex) of
        false -> digraph:add_vertex(Graph, Vertex);
        _ -> ok
    end.

check_edge(Graph, V1, V2, Label)->
    case edge(Graph, V1, V2) of
	[] -> digraph:add_edge(Graph, V1, V2, Label);
	_ -> ok
    end.

edge(G, V1, V2)->
    [digraph:edge(G, E)
     || E <- ?MISC:intersect(digraph:out_edges(G, V1),
                             digraph:in_edges(G, V2))].



cycles(Graph, [], Cycles, Edges)-> 
    [digraph:add_edge(Graph, E, V1, V2, Label) ||
	{E, V1, V2, Label} <- Edges],
    Cycles;
cycles(Graph, [Head|Tail] = Vertices, Cycles, Edges)->
    case digraph:get_cycle(Graph, Head) of
	false -> cycles(Graph, Tail, Cycles, Edges);
	[_  = Loop| []]  ->
	    cycles(Graph, Tail, [ [Loop, Loop] | Cycles], Edges);
	Cycle ->
	    [{E, _, _, _} = Edge | []] =
		edge(Graph, lists:nth(length(Cycle)-1, Cycle), Head),
	    digraph:del_edge(Graph, E),
	    cycles(Graph, Vertices, is_cycle(Cycles, Cycle,
					     lists:member(Cycle, Cycles)),
		   [ Edge| Edges])
    end.

is_cycle(Cycles, _, true)->
    Cycles;
is_cycle(Cycles, Cycle, false) ->
    [Cycle | Cycles].

build_subgraph(Graph, Node)->
    case get_funcall(Node) of
	[] -> ok;
	[{Fun, CalledList}]->  
	    case lists:subtract(CalledList, digraph:vertices(Graph)) of
		[] ->  build_vertex(Graph, [{Fun, CalledList}]);
		NewVertices -> build_vertex(Graph, [{Fun, CalledList}]),
		              [build_subgraph(Graph, [Called])|| Called<-NewVertices]
	    end
    end.



 


