%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös  Loránd University are  Copyright 2007-2009,
%%% Eötvös Loránd University. All Rights Reserved.

%%% @doc Functionblock relationship examiner.
%%%
%%% @author Kinga Szava <guca04@gmail.com>

-module(refusr_fb_relations).
-vsn("$Rev: 7223 $ ").

-export([get_relations/0, get_relations/1, get_relations/2, is_relation/1, check_cycle/0, check_cycle/1]).

-export([draw/0, draw_cycles/0, draw/1]).

-export([build_table/3]).



-include("user.hrl").

-record(row, {number,
	      element,
	      relations,
	      cycles = []
	     }).

-define(LibRel, refusr_lib_relations).


%% =============================================================================
%% Making proper input list

sort([])->
    case refusr_dir_sort:sort() of
        {error, _} -> throw(no_mod);
        List -> List
    end;
sort(List) when is_list(List)->
    sort_dir(List);
sort(_)->
    throw(badarg).

sort_dir([{_, _} | _]  = RelList) ->
    sort_dir(make_dir(RelList));
sort_dir(List) when is_list(List) ->
    case refusr_dir_sort:sort() of
        {error, _} -> throw(no_mod);
        DirSort -> [{Path, ModList} || {Path, ModList} <- DirSort,
                                       lists:member(Path, List)]
    end;
sort_dir(_) ->
    throw(badarg).


make_dir([])->
    [];
make_dir([{Base, FbList} | Rest])->
    [filename:join(Base, dir_to_list(Dir)) || Dir <- FbList] ++ make_dir(Rest).

dir_to_list(Dir) when is_integer(Dir) -> integer_to_list(Dir);
dir_to_list(Dir) when is_atom(Dir)    -> atom_to_list(Dir);
dir_to_list(_)                        -> throw(badarg).


get_node({_, _, _} = Node)->
	Node;
get_node(Name)->
    ?Query:exec(?Mod:find(Name)).


get_mod_funcall(Mod)->
    lists:flatten([Funcalls
                   || Fun <-[?Query:exec(Mod, ?Mod:locals())],
                      Funcalls <- [?Query:exec(Fun, ?Fun:funcalls())],
                      Funcalls /= []]).

get_func_mod(Fun)->
    ?Query:exec1(Fun, ?Fun:module(), mod_not_found).

%%% ============================================================================
%%% Building the dets table
%%
%% @spec build_table(Name, Args, Data) -> ok
%% @doc Builds up the proper dets table from a built digraph.
build_table(Name, Args, {undefined, _})->
	build_table(Name, Args, {sort([]), false});
build_table(Name, Args, {[], _})->
	build_table(Name, Args, {sort([]), false});
build_table(Name, _, Data)->
    Graph = digraph:new(),
    try
        build_graph(Graph, Data),
        Cycles = cycles(Graph, digraph:vertices(Graph), [], []),
	?LibRel:build_paths(Graph, Name, Cycles, fb_rel)
    after
        digraph:delete(Graph)
    end.

 

%%==============================================================================
%% Exported functions

%% @spec get_relations() -> [Relation_Pairs] | {error, bad_argument}
%% @doc Prints out every connection among every functionblock list.
%% Equivalent to get_relations([], true), see {@link get_relations/2}.
get_relations()->
     get_relations([], true).

%% -----------------------------------------------------------------------------
%% @spec get_relations(Arg) -> [Relation_Pairs] | {error, bad_argument}
%%	     Arg = Other::bool() | ParList
%%           ParList = [string()]
%%                   | [{Basename::string(), [FunctionBlock::atom()]}]
%%           Relation_Pairs = tuple(Path)
%%           Path = string()
%%	     Other = bool()
%% @doc If the argument is a boolean value, then it is equivalent to get_relations([], Other).
%% Otherwise, same as get_relations(ParList, true), see {@link get_relations/2}.
get_relations(Other) when not is_list(Other)->
	get_relations([], Other);
get_relations(ParList)->
	get_relations(ParList, true).

%% -----------------------------------------------------------------------------
%% @spec get_relations(ParList, Other) -> [Relation_Pairs] | {error, bad_argument}
%%           ParList = [string()]
%%                   | [{Basename::string(), [FunctionBlock::atom()]}]
%%           Relation_Pairs = tuple(Path)
%%           Path = string()
%%	     Other = bool()
%%
%% @doc Displays the relationship among elements of the given
%% functionblock list. There is a relationship between `fb1' block and
%% `fb2' block if a module of `fb1' is dependent of a module of
%% `fb2'. The result is a tuple-list whose elements represent the
%% relations.
%% The Other parameter stands for whether the category "Other" would be
%% taken into consideration or not. The default value is true.
get_relations({Path, [Hd | _] = List}, Other) when is_atom(Hd)->
	ParList = [filename:join(Path, Next)|| Next <- List], 
	get_relations(ParList, Other);
get_relations(ParList, Other)->
    io:format("~p~n", [get_rel(
			  ?LibRel:invest_table(fb_rel,
			 [{type, get_rel} | ParList]),
			[], Other)]).


%% -----------------------------------------------------------------------------
%% @spec is_relation(ParList)-> true | false | Error
%%           ParList = tuple(Fb)
%%	     Fb = string()
%%                   | {Basename::string(), [FunctionBlocks::atom()]}
%%           Error = {error, no_list_given}
%%                 | {error, bad_argument_list_length}
%%                 | {error, bad_argument_form}
%%
%% @doc Decides whether there is a connection between the two given
%% functionblocks.
is_relation([])->
    {error, no_list_given};
is_relation([{_, _} = Arg | []])->
    is_relation(tuple_to_list(Arg));
is_relation({_,_} = Arg)->
    is_relation(tuple_to_list(Arg));
is_relation(Arg) when not is_list(Arg) and not is_tuple(Arg)->
    {error, bad_argument_form};
is_relation(Arg) when length(Arg) == 2 ->
	case get_rel(?LibRel:invest_table(fb_rel, [{type, get_rel},
			   {gnode, Arg}]), [], true) of
		{error, Msg} -> {error, Msg};
		GetRel ->
   			lists:member(list_to_tuple(Arg), GetRel) or
			lists:member(list_to_tuple(lists:reverse(Arg)), GetRel)
   	 end;
is_relation(_) -> {error, bad_argument_list_length}.


%% -----------------------------------------------------------------------------
%% @spec check_cycle()-> {ok, no_cyclic_dependency}
%%                     | {integer(), list()}
%%                     | {error, bad_argument}
%%
%% @doc Looks for cycles among every function block list. Equivalent
%% to check_cycle([]). See {@link check_cycle/1}.
check_cycle()->
    check_cycle([]).

%% -----------------------------------------------------------------------------
%% @spec check_cycle(ParList) -> {ok, no_cyclic_dependency}
%%                             | {integer(), list()}
%%                             | {error, bad_argument}
%%           ParList = [string()]
%%                   | [{Basename::string(), [FunctionBlock::atom()]}]
%%
%% @doc Checks for cycles in the dependencies between the given
%% functionblock list.
check_cycle(ParList)->
   ?LibRel:invest_table(fb_rel, [{type, check}, {gnode, ParList}]).


%% =============================================================================
%% Draw graph

%% -----------------------------------------------------------------------------
%% @spec draw()-> {ok, no_cyclic_dependency} | {integer(), list()}
%% @doc Prints out the entire graph. Output file is fb_relations.dot.
draw()->
    draw([{gnode, []}]).

%% -----------------------------------------------------------------------------
%% @spec draw(PropList) -> {ok, no_cyclic_dependency}
%%                      | {integer(), list()}
%%                      | {error, bad_function_name}
%%
%% @doc Creates a subgraph drawing from the given functionblock list.
%% The file's name: fb_relations.dot or user defined.
%% PropList:
%% ```
%% {dot, DotFileName::string()}
%%
%% '''
%% User defined dot name.
%% ```
%% {gnode, [string()] | [{Basename::string(), [FunctionBlock::atom()]}]}
%% '''
%% Which function blocks should be inspected.
draw(PropList)->
    ?LibRel:invest_table(fb_rel, [{type, draw} | PropList]).

%% -----------------------------------------------------------------------------
%% @spec draw_cycles() -> {ok, no_cyclic_dependency}
%%                      | {integer(), list()}
%%                      | {error, bad_function_name}
%%
%% @doc Prints out a subgraph which contains the cycle(s). Unless
%% cycles exist, calls {@link draw/0}. Output file is
%% fb_relations.dot.
draw_cycles()->
    draw_cycles([{gnode, []}]).

%% -----------------------------------------------------------------------------
%% @spec draw_cycles(PropList) -> {ok, no_cyclic_dependency}
%%                      | {integer(), list()}
%%                      | {error, bad_function_name}
%%
%% @doc From the given function block list creates a graph from the cycle list.
%% If there are no cycles, draws out the entire graph.
%% The file's name: fb_relations.dot or user defined.
%% PropList:
%% ```
%% {dot, DotFileName::string()}
%%
%% '''
%% User defined dot name.
%% ```
%% {gnode, [string()] | [{Basename::string(), [FunctionBlock::atom()]}]}
%% '''
%% Which function blocks should be inspected.
draw_cycles(PropList)->
     ?LibRel:invest_table(fb_rel, [{type, cycles} | PropList]).


%%======================================================================

get_rel([], Result, _)->
    Result;
get_rel([{Dir, Calls} | Rest], Result, Other)->
    get_rel(Rest, divide_rel(Dir, Calls, Result, Other), Other);
get_rel([#row{number=_, element=Dir, relations = Calls, cycles=_} | Rest], Result, Other)->
    get_rel(Rest, divide_rel(Dir, Calls, Result, Other), Other);
get_rel([[#row{number=_, element=Dir, relations = Calls, cycles=_} | _] | Rest], Result, Other)->
    get_rel(Rest, divide_rel(Dir, Calls, Result, Other), Other).

divide_rel(_, [], Result, _)->
    Result;
divide_rel("Other", _, Result, false)->
    Result;
divide_rel(Dir, ["Other"| Rest], Result, false)->
    divide_rel(Dir, Rest, Result, false);
divide_rel(Dir, [Dir | Rest], Result, Other) ->
    divide_rel(Dir, Rest, Result, Other);
divide_rel(Dir, [Hd | Rest], Result, Other) ->
    Member = 
	lists:member({Dir, Hd}, Result) or
	lists:member({Hd, Dir}, Result),
    if
	Member == false->
	    divide_rel(Dir, Rest,  [{Dir, Hd} | Result], Other);
	true ->
	    divide_rel(Dir, Rest, Result, Other)
    end.

%%=============================================================
%% Graph functions
build_graph(Graph, {FbList, false})->
    [build_vertex(Graph, Element) || Element <- FbList];
build_graph(Graph, {Fbs, true})->
    build_node(Graph, Fbs, Fbs),
	[digraph:del_vertex(Graph, V)
	   || V<-digraph:vertices(Graph), 
		element(2, V) == module].

build_node(_, [], _)->
	ok;
build_node(Graph, [{Path, ModList} | Rest], Fbs)->
    check_vertex(Graph, Path),
    [begin
	check_vertex(Graph, Mod),
	check_edge(Graph, Path, Mod),
	check_called(Graph, {Path, Mod}, get_called_mods(Mod), Fbs)
     end|| Mod<-ModList],
    build_node(Graph, Rest, Fbs).

check_called(_, _, [], _)->
	ok;
check_called(Graph, {Path1, Mod}, [Hd|Rest], Fbs)->
     case is_member(Hd, Fbs) of
	{Path2, _} ->
		check_vertex(Graph, Hd),
		check_vertex(Graph, Path2),
		check_edge(Graph, Path2, Hd),
		check_edge(Graph, Path1, Path2); %%for Label
	false ->
		ok
    end,
    check_called(Graph, {Path1, Mod}, Rest, Fbs).

is_member(_, [])->
	false;
is_member(Mod, [{_, ModList} =Node | Rest])->
	case lists:member(Mod, ModList) of
		false ->
			is_member(Mod, Rest);
		true ->
			Node
	end.

build_vertex(Graph, {Path, ModList})->	
    check_vertex(Graph, Path),
    [begin
    	 check_vertex(Graph, CalledPath),
	 check_edge(Graph, Path, CalledPath)
     end || Mod <- ModList,
	    CalledPath <- get_called_dir(get_node(Mod))].

check_vertex(Graph, Vertex)->
    case  digraph:vertex(Graph, Vertex) of
        false -> digraph:add_vertex(Graph, Vertex);
        _ -> ok
    end.

check_edge(Graph, Vertex1, Vertex2)->
    case edge(Graph, Vertex1, Vertex2) of
	[] -> digraph:add_edge(Graph, Vertex1, Vertex2);
	_ -> ok
    end.

edge(Graph, Vertex1, Vertex2)->
    [digraph:edge(Graph, Edge)
     || Edge <- ?MISC:intersect(digraph:out_edges(Graph, Vertex1),
				digraph:in_edges(Graph, Vertex2))].

get_called_mods(Mod)->
    [get_func_mod(Called) || Called <- get_mod_funcall(Mod)].

get_called_dir(Mod)->
    [Path || Called <- get_mod_funcall(Mod),
             (CalledMod = get_func_mod(Called)) /= Mod,
             {Path, _} <- refusr_dir_sort:get_dirs([CalledMod])].

cycles(Graph, [], Cycles, Edges)-> 
    [digraph:add_edge(Graph, E, V1, V2, Label) ||
	{E, V1, V2, Label} <- Edges],
    Cycles;
cycles(Graph, [Head|Tail] = Vertices, Cycles, Edges)->
    case digraph:get_cycle(Graph, Head) of
	false -> cycles(Graph, Tail, Cycles, Edges);
	[_  = Loop| []]  ->
	    cycles(Graph, Tail, [ [Loop, Loop] | Cycles], Edges);
	Cycle ->
	    [{E, _, _, _} = Edge | []] =
		edge(Graph, lists:nth(length(Cycle)-1, Cycle), Head),
	    digraph:del_edge(Graph, E),
	    cycles(Graph, Vertices, is_cycle(Cycles, Cycle,
					     lists:member(Cycle, Cycles)),
		   [ Edge| Edges])
    end.

is_cycle(Cycles, _, true)->
    Cycles;
is_cycle(Cycles, Cycle, false) ->
    [Cycle | Cycles].

